!=========================================================================!
! file pwcr_sub.f90                                                       !
!=========================================================================!
!      IMPROVED CYCLIC REDUCTION FOR SOLVING QUEUEING PROBLEMS            !
!                     by D.A. Bini and B. Meini                           !
!               (bini@dm.unipi.it    meini@dm.unipi.it)                   !
!           Fortran 90 Program version 1.0, January 30, 1997              !
!                          SUBROUTINES                                    !
!=========================================================================!
! The following subroutines constitute a package for the solution of      !
! the matrix equation                                                     !
!                A_1+X A_2+X^2 A_3+....X^(m-1) A_m-X=O             (1)    !
! where X is the (nb x nb) unknown matrix and the (nb x nb)               !
! nonnegative matrices A_i, i=1,2,..., are such that  A_1+A_2+...+A_m     !
! is column stochastic.                                                   !
! The method used in the subroutines is based on the cyclic reduction     !
! technique expressed in functional form and generates a sequence of      !
! approximations converging quadratically to the solution of (1).         !
! Unlike the method of Bini-Meini, SIMAX 1996, here the cyclic            !
! reduction step is implemented in a point-wise style. That is, all       !
! the intermediate matrix power series are evaluated at the set of        ! 
! Fourier points. This allows us to keep to the minimum value the         !
! size of the Fourier transforms involved in the subroutines with a       !
! consequent improvement of the performance of our algorithm.             !
!=========================================================================!
! This package is made up by the following subroutines                    !
!  pwcr (Point-Wise Cyclic Reduction): approximate the solution of        !
!             (1) by means of cyclic reduction.                           !
!  computeG : approximate the solution of (1) if m=2.                     !
!  residual : compute the 1-norm of the residual                          !
!             G-A_1+G A_2+G^2 A_3+....G^(m-1) A_m,                        !
!             where G is an approximation of the solution X of (1).       !
!  schur    : execute one step of point-wise cyclic reduction by          !
!             computing the Schur complement (formulae (7) and (8) in the !
!             paper)                                                      !
!  test     : auxiliary subroutine used by schur                          !
!  sc1p     : auxiliary subroutine used by schur                          !
!  sc2p     : auxiliary subroutine used by schur                          !
!  scc2p    : auxiliary subroutine used by schur                          !
!  sc1      : auxiliary subroutine used by schur                          !
!  sc2      : auxiliary subroutine used by schur                          !
!  scc2     : auxiliary subroutine used by schur                          !
!  prodc    : compute complex matrix product by performing 3 real         !
!             matrix multiplications.                                     !
!  means    : auxiliary subroutine used by test                           !
!  pmeans   : auxiliary subroutine used by test                           !
!  solver   : auxiliary subroutine for solving real linear systems        !
!  solvec   : auxiliary subroutine for solving complex linear systems     !
!=========================================================================!
! This package makes use also of auxiliary routines, for performing       !
! FFT computation, which have been collected in the separate file         !
! pwcr_fft.f90;  the requested interfaces have been collected in the file !
! fft_int.f90.                                                            !
! Moreover,  LAPACK and LAPACK90 subroutines and interfaces are used:     !
! the subroutines have been collected in the separate files solve.f and   !
! kgesv1.f90; the interfaces are contained in the files laauxmod.f90,     !
! laf77mod.f90, laf90mod.f90                                              !
!=========================================================================!

!=========================================================================!
!                          SUBROUTINE PWCR                                !
!=========================================================================!
! This subroutine computes the matrix X solving the equation              !
!                A_1+XA_2+X^2A_3+....X^(m-1)A_m-X=O                       !
! where A_1+A_2+...+A_m is a column stochastic kxk matrix,                !
! by means of the method of pointwise cyclic reduction                    !
!=========================================================================!
! Input varables:                                                         !
!    a   : pointer associated with the blocks A_i,  i=1,m                 !
!    eps : error bound used for checking the stop condition at each       !
!          step of the cyclic reduction.                                  !
!=========================================================================!
! Output variables:                                                       !
!    g   : pointer associated with the approximation of the solution X    !
!          of (1)                                                         ! 
!    err : 1-norm of the left-hand side of (1) where X is replaced by  g  !
!=========================================================================!
! Interfaces used: schur_interface, pwcr_interface                        !
!=========================================================================!
! Subroutines used: schur, computeG, residual, pmeans                     !
!=========================================================================!


SUBROUTINE PWCR(a,eps,g,err)
  USE schur_interface
  USE pwcr_interface, ONLY : computeG, residual
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:,:), POINTER :: a, ae, ao, hae, hao
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER   :: a1, g
  REAL(KIND(0.d0)), DIMENSION(:), POINTER     :: mean, hmean, a1s
  REAL(KIND(0.d0))                            :: err, eps, stoc
  INTEGER                                     :: m, n, nb, j, &
       l, l1, ln, step
  INTRINSIC size 
  !-------------------------------------
  ! Prepare the input
  !-------------------------------------
  m=SIZE(a,3)
  nb=SIZE(a,1)
  l1=(m+1)/2
  n=l1
  ln=LOG(1.0d0*n)/LOG(2.0d0)
  IF(2**ln<n)n=2**(ln+1)
  ALLOCATE(ae(nb,nb,n))
  ALLOCATE(ao(nb,nb,n))
  ALLOCATE(hae(nb,nb,n))
  ALLOCATE(hao(nb,nb,n))
  ALLOCATE(a1(nb,nb))
  ALLOCATE(a1s(nb))
  ae=0.0d0
  ao=0.0d0
  hae=0.0d0
  hao=0.0d0
  loop1 : DO l=1,l1-1 
     ao(:,:,l)=a(:,:,2*l)
     ae(:,:,l)=a(:,:,2*l-1)
  END DO loop1
  IF(2*l1==m)THEN  
     ao(:,:,l1)=a(:,:,m)
     ae(:,:,l1)=a(:,:,m-1)
  ELSE 
     ae(:,:,l1)=a(:,:,m)
  ENDIF
  DO l=1,nb
     stoc=SUM(ao(:,l,:))+SUM(ae(:,l,:))
     IF(1.0-stoc>eps)THEN
        WRITE(*,*)"The matrix is not stochastic"
        STOP
     ENDIF
  END DO
  hae(:,:,1:n-1)=ae(:,:,2:n)
  hao=ao
  a1(:,:)=ae(:,:,1)
  loop6 : DO j=1,nb
     a1s(j)=SUM(a1(:,j))
  END DO loop6
  !---------------------------------
  ! Start Pointwise Cyclic Reduction
  !---------------------------------
  step=0
  n=SIZE(ae,3)
  pwcrstage : DO WHILE(n>1) 
     step=step+1
     CALL pmeans(ae,ao,hae,hao,mean,hmean)
     CALL schur(ae,ao,hae,hao,mean,hmean,a1s,eps)
     n=SIZE(ae,3)
  END DO pwcrstage
  CALL computeG(a1,a1s,ae,ao,hae,hao,eps,g)
  CALL residual(a,g,err)
  RETURN
END SUBROUTINE PWCR



!=========================================================================!
!                          SUBROUTINE COMPUTEG                            !
!=========================================================================!
! This subroutine computes the matrix X solving the equation              !
!                A_1+XA_2+X^2A_3+....X^(m-1)A_m-X=O                       !
! in the case where m=2.                                                  !
!=========================================================================!
! Input variables:                                                        !
!    a1  : pointer associated with the block A_1                          !
!    a1s : pointer associated with the vector (1,1,...,1)A1               !
!    ae  : pointer associated with the block coefficients of the series   !
!          \phi_{even}, i.e.,  A_1, A_3, ...                              !
!    ao  : pointer associated with the block coefficients of the series   !
!          \phi_{odd}, i.e,   A_2, A_4, ...                               !
!    hae : pointer associated withthe block coefficients of the series    !
!          \hat\phi_{even}                                                !
!    hao : pointer associated with the block coefficients of the series   !
!          \hat \phi_{odd}                                                !
!    eps : error bound used for checking the stop condition.              !
!=========================================================================!
! Output variables:                                                       !
!    g   : pointer associated with the  approximation of  the solution    !
!          X of (1)                                                       ! 
!=========================================================================!
! Interfaces used: lapack90_interfaces                                    !
!=========================================================================!
! Subroutines used: la_gesv                                               !
!=========================================================================!
SUBROUTINE computeG(a1,a1s,ae,ao,hae,hao,eps,g)
  USE lapack90_interfaces
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:,:), POINTER :: ae, ao, hae, hao
  REAL(KIND(0.d0)), DIMENSION(:,:),POINTER    :: a1, g, r
  REAL(KIND(0.d0)), DIMENSION(:), POINTER     :: a1s
  REAL(KIND(0.d0))                            :: s, eps
  INTEGER, DIMENSION(:), POINTER              :: ipiv
  INTEGER                                     :: n, nb, i, info
  LOGICAL :: stoc
  INTRINSIC size
  nb=SIZE(ae,1)
  n=SIZE(ae,3)
  ALLOCATE(g(nb,nb))
  ALLOCATE(ipiv(nb))
  stoc=.TRUE.
  loop50 : DO i=1,nb
     s=a1s(i)+SUM(hao(:,i,1))
     IF(1.0d0-s>eps)stoc=.FALSE.
  END DO loop50
  IF(stoc)THEN
     g(:,:)=-TRANSPOSE(hao(:,:,1))
     loop10 : DO i=1,nb
        g(i,i)=g(i,i)+1.0d0
     END DO loop10
     a1=TRANSPOSE(a1)
     CALL la_gesv(g, a1,ipiv, info)
     g=TRANSPOSE(a1)
     RETURN
  ENDIF
  ALLOCATE(r(nb,nb))
  g=-TRANSPOSE(ao(:,:,1))
  loop100 : DO i=1,nb
     g(i,i)=g(i,i)+1.0d0
  END DO loop100
  r=TRANSPOSE(ae(:,:,1))
  CALL la_gesv(g, r, ipiv, info)
  r=TRANSPOSE(r)
  g=-hao(:,:,1)-MATMUL(hae(:,:,1),r)
  g=TRANSPOSE(g)
  loop200 : DO i=1,nb
     g(i,i)=g(i,i)+1.0d0
  END DO loop200
  a1=TRANSPOSE(a1)
  CALL la_gesv(g, a1, ipiv, info)  
  g=TRANSPOSE(a1) 
  RETURN
END SUBROUTINE computeG

!=========================================================================!
!                          SUBROUTINE RESIDUAL                            !
!=========================================================================!
! This subroutine computes the residual ERR, i.e., the 1-norm             !
!             ERR = || A_1+GA_2+G^2A_3+....G^(m-1)A_m-G ||                !
! where G is an approximation of the solution X of (1)                    !
!=========================================================================!
! Input variables:                                                        !
!    a   : pointer associated with the blocks A_1, A_2,..., A_m           !
!    g   : pointer associated with the approximation of the solution      !
!        X of (1)                                                         !
!=========================================================================!
! Output variables:                                                       !
!    err : the seeked residual                                            ! 
!=========================================================================!
SUBROUTINE residual(a,g,err)
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:,:), POINTER :: a
  REAL(KIND(0.d0)), DIMENSION(:,:), POINTER   :: g, r
  REAL(KIND(0.d0))                            :: err, s
  INTEGER                                     :: n, nb, i
  INTRINSIC size 
  nb=SIZE(a,1)
  n=SIZE(a,3)
  ALLOCATE(r(nb,nb))
  r=a(:,:,n)
  loop1 : DO i=n-1,1,-1
     r=MATMUL(g,r)+a(:,:,i)
  END DO loop1
  r=ABS(r-g)
  err=0.0d0
  loop2 : DO i=1,nb
     s=SUM(r(:,i))
     IF(s>err) err=s
  END DO loop2
  RETURN
END SUBROUTINE residual

!=========================================================================!
!                          SUBROUTINE SCHUR                               !
!=========================================================================!
! This subroutine performs one step of cyclic reduction by computing      !
! the even and the odd components of the matrix series \phi^{(n+1)} and   !
! \hat\phi^{(n+1)} (compare formulae (7), (8) in the paper) defining the  !
! Schur complement in the cyclic reduction process.                       !
! The subroutine :                                                        !
! - interpolates the series \phi_{even}, \phi_{odd}, \hat\phi_{even},     !
!   \hat\phi_{odd}, numerically truncated to polynomials of degree n-1,   !
!    at the n-th roots of 1,                                              !
! - performes convolutions according to formula (7) of the paper          !
! - computes the matrix polynomials interpolating these values            !
! - checks the accuracy of the result; if the result is not accurate then !
!   doubles the number of interpolation points and repeats; otherwise     !
!   outputs the block coefficients of the matrix power series.            !
!=========================================================================!
! Input variables:                                                        !
!    ae  : pointer associated with the block coefficients of the series   !
!          \phi_{even}, i.e.,  A_1, A_3, ...                              !
!    ao  : pointer associated with the block coefficients of the series   !
!          \phi_{odd}, i.e,   A_2, A_4, ...                               !
!    hae : pointer associated withthe block coefficients of the series    !
!          \hat\phi_{even}                                                !
!    hao : pointer associated with the block coefficients of the series   !
!          \hat \phi_{odd}                                                !
!    mean :  pointer associated with the vector (1,...,1)\phi'(1),        !
!            for the current function \phi                                !
!    hmean:  pointer associated with the vector (1,...,1)\hat\phi'(1) for !
!            the current function   \hat\phi                              !
!    a1s  :  pointer associated with the vector (1,...,1)\phi(0), for     !
!            the initial function \phi                                    !
!    eps  : error bound used for checking the stop condition.             !
!=========================================================================!
! Output variables: new values for AE, AO, HAE, HAO, MEAN, HMEAN          !
!=========================================================================!
! Interfaces used:   fft_interface, schur_interface                       !
!=========================================================================!
! Subroutines used: ftb1, iftb1, ftb2, iftb2, sc1p, sc2p, scc2p, sc1, sc2 !
!           scc2, means, test                                             !
!=========================================================================!
SUBROUTINE schur(ae,ao,hae,hao,mean,hmean,a1s,eps)
  USE fft_interface
  USE schur_interface, ONLY : sc1, sc2p, sc1p, sc2, scc2p, scc2, &
       means, test
  IMPLICIT NONE
  REAL(KIND(0.d0)), DIMENSION(:,:,:), POINTER :: ae, ao, hae, hao
  REAL(KIND(0.d0)), DIMENSION(:), POINTER     :: mean, hmean, a1s
  REAL(KIND(0.d0))                            :: eps
  REAL(KIND(0.d0)), DIMENSION(:,:,:), POINTER :: tae, tao, thae, thao, &
       a1, a2, a3, a4  
  REAL(KIND(0.0d0))                           :: s1, s2
  INTEGER                                     :: l,nn, nb, n, nradd
  LOGICAL                                     :: answer, hanswer
  INTRINSIC size
  nb=SIZE(ae,1)
  n=SIZE(ae,3)
  ALLOCATE(tae(nb,nb,n))
  ALLOCATE(tao(nb,nb,n))
  ALLOCATE(a1(nb,nb,n))
  ALLOCATE(a2(nb,nb,n))
  CALL ftb1(ae,tae)
  CALL ftb1(ao,tao)
  CALL sc1p(tae,tao,a1)
  CALL sc2p(tae,tao,a1,a2)
  ALLOCATE(a3(nb,nb,n))
  CALL iftb1(a2,a3)
  ALLOCATE(thae(nb,nb,n))
  ALLOCATE(thao(nb,nb,n))
  CALL ftb1(hae,thae)
  CALL ftb1(hao,thao)
  CALL scc2p(thae,thao,a1,a2)
  ALLOCATE(a4(nb,nb,n))
  CALL iftb1(a2,a4)
  CALL means(tae,tao,thae,mean,hmean)
  nn=n
  nradd=0
100 nradd=nradd+1
  ! check the accuracy of the matrix coefficients
  CALL test(a3,mean,eps,answer)
  CALL test(a4,hmean,eps,hanswer)
  IF(answer.OR.hanswer) THEN ! double the number of interpolation points
     IF(nradd>1)THEN
        n=nn/2
        a1=ae
        DEALLOCATE(ae)
        ALLOCATE(ae(nb,nb,nn))
        ae(:,:,1:n)=a1
        ae(:,:,n+1:nn)=0.0d0
        a1=ao
        DEALLOCATE(ao)
        ALLOCATE(ao(nb,nb,nn))
        ao(:,:,1:n)=a1
        ao(:,:,n+1:nn)=0.0d0
        a1=hae
        DEALLOCATE(hae)
        ALLOCATE(hae(nb,nb,nn))
        hae(:,:,1:n)=a1
        hae(:,:,n+1:nn)=0.0d0
        a1=hao
        DEALLOCATE(hao)
        ALLOCATE(hao(nb,nb,nn))
        hao(:,:,1:n)=a1
        hao(:,:,n+1:nn)=0.0d0
        DEALLOCATE(a1)
     ENDIF
     CALL ftb2(ae,tae)
     CALL ftb2(ao,tao)
     CALL sc1(tae,tao,a1)
     CALL sc2(tae,tao,a1,a2)
     CALL iftb2(a2,a3)
     CALL ftb2(hae,thae)
     CALL ftb2(hao,thao)
     CALL scc2(thae,thao,a1,a2)
     CALL iftb2(a2,a4)
     nn=2*nn
     GOTO 100 
  END IF
  DEALLOCATE(ae)
  DEALLOCATE(ao)
  DEALLOCATE(hae)
  DEALLOCATE(hao)
  n=nn/2
  ALLOCATE(ae(nb,nb,n))
  ALLOCATE(ao(nb,nb,n))
  ALLOCATE(hae(nb,nb,n))
  ALLOCATE(hao(nb,nb,n))
  loop1 : DO l=1,nn/2
     ae(:,:,l)=a3(:,:,2*l-1)
     ao(:,:,l)=a3(:,:,2*l)
     hae(:,:,l)=a4(:,:,2*l)
     hao(:,:,l)=a4(:,:,2*l-1)
  END DO loop1
  loop2 : DO l=1,nb
     s1=a1s(l)+SUM(hae(:,l,:))+SUM(hao(:,l,:))
     s2=SUM(ae(:,l,:))+SUM(ao(:,l,:))
     s1=1.d0/s1
     s2=1.d0/s2
     ae(:,l,:)=ae(:,l,:)*s2
     ao(:,l,:)=ao(:,l,:)*s2
     hae(:,l,:)=hae(:,l,:)*s1
     hao(:,l,:)=hao(:,l,:)*s1
  END DO loop2
  RETURN
END SUBROUTINE schur


!=========================================================================!
!                          SUBROUTINE TEST                                !
!=========================================================================!
! This subroutine applies the test (12)  in the paper and outputs         !
! answer=.true. if the number of interpolation point must be doubled      !
! (failure of the test), otherwise outputs answer=.false.                 !
!=========================================================================!
! Input variables                                                         !
!   u   :  pointer associated with the matrix coefficients of the         !
!          polynomial interpolating \phi (or \hat\phi) at the root of 1.  !
!   mean:  pointer associated with the exact value of (1,...,1)\phi'(1),  !
!          or (1,...,1)\hat\phi'(1),                                      !
!   eps :  error bound                                                     !
!=========================================================================!
! Output variables                                                        !
!   answer                                                                !
!=========================================================================!
SUBROUTINE test(u, mean, eps, answer)
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: u
  REAL(KIND(0.d0)), POINTER, DIMENSION(:)     :: mean
  REAL(KIND(0.d0))                            :: eps, v, err
  INTEGER                                     :: j, l, n, nb
  LOGICAL                                     ::  answer
  INTRINSIC size
  n=SIZE(u,3)
  nb=SIZE(u,1)
  answer=.FALSE.
  err=0.d0
  loop10 : DO  j=1,nb
     v=0.d0
     loop20 : DO  l=2,n
        v=v+SUM(u(:,j,l))*(l-1)
     END DO loop20
     IF(ABS(v-mean(j))>eps)answer=.TRUE.
     IF(err<ABS(v-mean(j)))err=ABS(v-mean(j))
  END DO loop10
  RETURN
END SUBROUTINE test



!=========================================================================!
!                          SUBROUTINE SC1P                                !
!=========================================================================!
! This subroutine computes the values of the function                     !
!                     \phi_{even}(I-\phi_{odd})^{-1}                      !
! (compare formula (8) of the paper) at the n-th root of 1, where         !
! the matrix series \phi_{even}, \phi_{odd} are numerically truncated     !
! to polynomials of degree n-1.                                           !
!=========================================================================!
! Input variables                                                         !
!   tae  :  pointer associated with the matrix values of \phi_{even} at   !
!           the roots of 1                                                !
!   tao  :  pointer associated with the matrix values of \phi_{odd} at    !
!           the roots of 1                                                !
!=========================================================================!
! Output variables                                                        !
!   a    :  pointer associated with the matrix values of                  !
!           \phi_{even}(I-\phi_{odd})^{-1} at the roots of 1              !
!=========================================================================!
! Used interfaces:  schur_interface                                       !
!=========================================================================!
! Used subroutines: solver, solvec                                        !
!=========================================================================!
SUBROUTINE sc1p(tae,tao,a)
  USE schur_interface, ONLY : solver, solvec
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: tao, tae, a
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:)   :: rmat, termr
  COMPLEX(KIND(0.d0)), POINTER, DIMENSION(:,:):: cmat, termc
  INTEGER                                     :: i, n, nb, l, nm1
  INTRINSIC size
  n=SIZE(tae,3)
  nb=SIZE(tae,1)
  ALLOCATE(rmat(nb,nb))
  ALLOCATE(termr(nb,nb))
  ALLOCATE(cmat(nb,nb))
  ALLOCATE(termc(nb,nb))
  rmat(:,:)=-tao(:,:,1)
  loopi1 : DO i=1,nb
     rmat(i,i)=rmat(i,i)+1.0d0
  END DO loopi1
  termr(:,:)=tae(:,:,1)
  CALL solver(rmat,termr)
  a(:,:,1)=termr(:,:)
  nm1=n/2+1
  rmat(:,:)=-tao(:,:,nm1)
  loopi2 : DO i=1,nb
     rmat(i,i)=rmat(i,i)+1.0d0
  END DO loopi2
  termr(:,:)=tae(:,:,nm1)
  CALL solver(rmat,termr)
  a(:,:,nm1)=termr
  IF(n.EQ.2)THEN
     RETURN
  ENDIF
  loopl : DO l=1,n/2-1
     cmat(:,:)=-tao(:,:,l+1)-(0.0d0,1.0d0)*tao(:,:,n-l+1)
     loopi : DO i=1,nb
        cmat(i,i)=cmat(i,i)+1.0d0
     END DO loopi
     termc(:,:)=tae(:,:,l+1)+(0.0d0,1.0d0)*tae(:,:,n-l+1)
     CALL solvec(cmat,termc)
     a(:,:,l+1)=REAL(termc(:,:))
     a(:,:,n-l+1)=REAL((0.0d0,-1.0d0)*termc(:,:))
  END DO loopl
  RETURN
END SUBROUTINE sc1p


!=========================================================================!
!                          SUBROUTINE SC2P                                !
!=========================================================================!
! This subroutine computes the values of the function \phi in the         !
! left-hand side of (7) in the paper, at the n-th root of 1, where        !
! the matrix series \phi_{even}, \phi_{odd} are numerically truncated     !
! to polynomials of degree n-1.                                           !
!=========================================================================!
! Input variables                                                         !
!   tae :  pointer associated with the matrix values of \phi_{even} at    !
!          the roots of 1                                                 !
!   tao :  pointer associated with the matrix values of \phi_{odd} at     !
!          the roots of 1                                                 !
!   a1  : the value computed by the subroutine SC1P                       !
!=========================================================================!
! Output variables                                                        !
!   a2  :  pointer associated with the matrix values of the function      !
!          \phi of (7) in the paper, at the roots of 1                    !
!=========================================================================!
! Used interfaces:  schur_interface,   fft_interface                      !
!=========================================================================!
! Used subroutines: prodc, fillroots                                      !
!=========================================================================!
SUBROUTINE sc2p(tae,tao,a1,a2)
  USE schur_interface, ONLY : prodc
  USE fft_interface, ONLY   : fillroots
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: tao, tae, a1, a2
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:)   :: rmat, cmat
  REAL(KIND(0.0d0)), POINTER, DIMENSION(:)    :: wr, wi
  INTEGER                                     :: n, nb, l, nm1, &
       lm, mm, nmax
  COMMON wr, wi
  INTRINSIC size
  n=SIZE(tae,3)
  nb=SIZE(tae,1)
  ALLOCATE(rmat(nb,nb))
  ALLOCATE(cmat(nb,nb))
  rmat=MATMUL(a1(:,:,1),tae(:,:,1))
  a2(:,:,1)=tao(:,:,1)+rmat(:,:)
  nm1=n/2+1
  rmat=MATMUL(a1(:,:,nm1),tae(:,:,nm1))
  a2(:,:,nm1)=-tao(:,:,nm1)+rmat(:,:)
  IF(n.EQ.2)THEN
     RETURN
  ENDIF
  CALL fillroots(n)
  nmax=SIZE(wr,1)
  mm=nmax/n
  loopl : DO l=1,n/2-1
     lm=mm*l+1
     CALL prodc(a1,tae,l+1,n-l+1,rmat,cmat)
     a2(:,:,l+1)=wr(lm)*tao(:,:,l+1)-wi(lm)*tao(:,:,n-l+1)
     a2(:,:,l+1)=a2(:,:,l+1)+rmat(:,:)
     a2(:,:,n-l+1)=wr(lm)*tao(:,:,n-l+1)+wi(lm)*tao(:,:,l+1)
     a2(:,:,n-l+1)=a2(:,:,n-l+1)+cmat(:,:)
  END DO loopl
  RETURN
END SUBROUTINE sc2p


!=========================================================================!
!                          SUBROUTINE SCC2P                               !
!=========================================================================!
! This subroutine computes the values of the function \hat\phi in the     !
! left-hand side of (7) in the paper, at the n-th root of 1, where        !
! the matrix series \hat\phi_{even}, \hat\phi_{odd} are numerically       !
!truncated  to polynomials of degree n-1.                                 !
!=========================================================================!
! Input variables                                                         !
!   thae :  pointer associated with the matrix values of                  !
!           \hat\phi_{even} at the roots of 1                             !
!   thao :  pointer associated with the matrix values of \hat\phi_{odd}   !
!           at the roots of 1                                             !
!   a1   :  pointer associated with the value computed by the subroutine  !
!           SC1P                                                          !
!=========================================================================!
! Output variables                                                        !
!   a2  :  pointer associated with the matrix values of the function      !
!          \hat\phi of (7) at the roots of 1                              !
!=========================================================================!
! Used interfaces:  schur_interface                                       !
!=========================================================================!
! Used subroutines: prodc                                                 !
!=========================================================================!
SUBROUTINE scc2p(thae,thao,a1,a2)
  USE schur_interface, ONLY : prodc
  IMPLICIT NONE
  REAL(KIND(0.d0)),POINTER,DIMENSION(:,:,:) :: thao, thae, a1, a2
  REAL(KIND(0.d0)),POINTER,DIMENSION(:,:) :: rmat, cmat
  INTEGER :: n,nb, l, nm1
  INTRINSIC size
  n=SIZE(thae,3)
  nb=SIZE(thae,1)
  ALLOCATE(rmat(nb,nb))
  ALLOCATE(cmat(nb,nb))
  rmat=MATMUL(a1(:,:,1),thae(:,:,1))
  a2(:,:,1)=thao(:,:,1)+rmat(:,:)
  nm1=n/2+1
  rmat=MATMUL(a1(:,:,nm1),thae(:,:,nm1))
  a2(:,:,nm1)=thao(:,:,nm1)+rmat(:,:)
  IF(n.EQ.2)THEN
     RETURN
  ENDIF
  loopl : DO l=1,n/2-1
     CALL prodc(a1,thae,l+1,n-l+1,rmat,cmat)
     a2(:,:,l+1)=thao(:,:,l+1)+rmat(:,:)
     a2(:,:,n-l+1)=thao(:,:,n-l+1)+cmat(:,:)
  END DO loopl
  RETURN
END SUBROUTINE scc2p

!=========================================================================!
!                          SUBROUTINE SC1                                 !
!=========================================================================!
! This subroutine computes the values of the function                     !
!                     \phi_{even}(I-\phi_{odd})^{-1}                      !
! (compare formula (7) of the paper) at the odd-numbered roots of 1       !
!=========================================================================!
! Input variables                                                         !
!   tae  :  pointer associated with the matrix values of \phi_{even}      !
!           at the odd-numbered roots of 1                                !
!   tao  :  pointer associated with the matrix values of \phi_{odd}       !
!           at the odd-numbered roots of 1                                !
!=========================================================================!
! Output variables                                                        !
!   a    :  pointer associated with the matrix values of                  !
!           \phi_{even}(I-\phi_{odd})^{-1} at the odd-numbered roots of 1 !
!=========================================================================!
! Used interfaces:  schur_interface                                       !
!=========================================================================!
! Used subroutines:  solvec                                               !
!=========================================================================!
SUBROUTINE sc1(tae,tao,a)
  USE schur_interface, ONLY : solvec
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:)  :: tae, tao, a
  COMPLEX(KIND(0.d0)), POINTER, DIMENSION(:,:) :: cmat, termc
  INTEGER                                      :: n, nb, num, i, l
  INTRINSIC size
  n=SIZE(tae,3)
  nb=SIZE(tae,1)
  ALLOCATE(cmat(nb,nb))
  ALLOCATE(termc(nb,nb))
  num=n/2
  loopl : DO l=1,num
     cmat=-tao(:,:,l)-(0.0d0,1.0d0)*tao(:,:,num+l)
     loopi : DO i=1,nb
        cmat(i,i)=cmat(i,i)+1.0d0
     END DO loopi
     termc=tae(:,:,l)+(0.0d0,1.0d0)*tae(:,:,num+l)
     CALL solvec(cmat,termc)
     a(:,:,l)=REAL(termc(:,:))
     a(:,:,num+l)=REAL((0.0d0,-1.0d0)*termc(:,:))
  END DO loopl
  RETURN
END SUBROUTINE sc1

!=========================================================================!
!                          SUBROUTINE SC2                                 !
!=========================================================================!
! This subroutine computes the values of the function \phi in the         !
! left-hand side of (7) in the paper, at the odd-numbered roots of 1      !
!=========================================================================!
! Input variables                                                         !
!   tae  :  pointer associated with the matrix values of \phi_{even}      !
!           at the odd-numbered roots of 1                                !
!   tao  :  pointer associated with the matrix values of \phi_{odd}       !
!           at the odd-numbered roots of 1                                !
!   a1   : pointer associated with the value computed by the subroutine   !
!           sc1                                                           !
!=========================================================================!
! Output variables                                                        !
!   a2  : pointer associated with the matrix values of the function       !
!         \phi of (7) at the odd-numbered roots of 1                      !
!=========================================================================!
! Used interfaces:  schur_interface,   fft_interface                      !
!=========================================================================!
! Used subroutines: prodc, fillroots                                      !
!=========================================================================!
SUBROUTINE sc2(tae,tao,a1,a2)
  USE fft_interface, ONLY   : fillroots
  USE schur_interface, ONLY : prodc
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: tae, tao, a1, a2
  REAL(KIND(0.d0)), POINTER, DIMENSION(:)     :: wr, wi
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:)   :: rmat, cmat
  INTEGER                                     :: n, nb, num, l, lm, &
       mm, nmax
  COMMON wr, wi
  INTRINSIC size
  n=SIZE(tae,3)
  nb=SIZE(tae,1)
  ALLOCATE(cmat(nb,nb))
  ALLOCATE(rmat(nb,nb))
  num=n/2
  CALL fillroots(2*n)
  nmax=SIZE(wr,1)
  mm=nmax/(2*n)
  loopl : DO l=1,num
     lm=mm*(2*l-1)+1
     CALL prodc(a1,tae,l,num+l,rmat,cmat)
     a2(:,:,l)=wr(lm)*tao(:,:,l)-wi(lm)*tao(:,:,num+l)
     a2(:,:,l)=a2(:,:,l)+rmat(:,:)
     a2(:,:,num+l)=wr(lm)*tao(:,:,num+l)+wi(lm)*tao(:,:,l)
     a2(:,:,num+l)=a2(:,:,num+l)+cmat(:,:)
  END DO loopl
  RETURN
END SUBROUTINE sc2


!=========================================================================!
!                          SUBROUTINE SCC2                                !
!=========================================================================!
! This subroutine computes the values of the function \hat\phi in the     !
! left-hand side of (7) in the paper, at the odd-numbered  roots of 1     !
!=========================================================================!
! Input variables                                                         !
!   thae :  pointer associated with the matrix values of                  !
!           \hat\phi_{even} at the roots of 1                             !
!   thao :  pointer associated with the matrix values of \hat\phi_{odd}   !
!           at the roots of 1                                             !
!   a1   :  pointer associated with the value computed by the subroutine  !
!           SC1                                                           
!=========================================================================!
! Output variables                                                        !
!   a2  :   pointer associated with the matrix values of the function     !
!           \hat\phi of (7) at the  odd-numbered roots of 1               !
!=========================================================================!
! Used interfaces:  schur_interface                                       !
!=========================================================================!
! Used subroutines: prodc                                                 !
!=========================================================================!
SUBROUTINE scc2(thae, thao, a1, a2)
  USE schur_interface, ONLY : prodc
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: thae, thao, a1, a2
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:)   :: rmat, cmat
  INTEGER                                     :: n, nb, l, num
  INTRINSIC size
  nb=SIZE(thae,1)
  n=SIZE(thae,3)
  ALLOCATE(rmat(nb,nb))
  ALLOCATE(cmat(nb,nb))
  num=n/2
  loopl : DO l=1,num
     CALL prodc(a1,thae,l,num+l,rmat,cmat)
     a2(:,:,l)=thao(:,:,l)+rmat(:,:)
     a2(:,:,num+l)=thao(:,:,num+l)+cmat(:,:)
  END DO loopl
  RETURN
END SUBROUTINE scc2


!=========================================================================!
!                          SUBROUTINE PRODC                               !
!=========================================================================!
! This subroutine computes the matrix product                             !
!  rmat+I*cmat = (a(:,:,ir)+I*a(:,:,ic))* (b(:,:,ir)+I*b(:,:,ic))         !
! by using the algorithm that performs three real matrix multiplications  !    
!=========================================================================!
SUBROUTINE prodc(a, b, ir, ic, rmat, cmat)
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:,:) :: a, b
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:)   :: rmat, cmat,aux
  INTEGER                                     :: ir,ic,nb
  INTRINSIC size
  nb=SIZE(a,1)
  ALLOCATE(aux(nb,nb))
  aux(:,:)=a(:,:,ir)+a(:,:,ic)
  rmat(:,:)=b(:,:,ir)+b(:,:,ic)
  cmat=MATMUL(aux,rmat)
  aux=MATMUL(a(:,:,ic),b(:,:,ic))
  rmat=MATMUL(a(:,:,ir),b(:,:,ir))
  cmat=cmat-rmat-aux
  rmat=rmat-aux
  RETURN
END SUBROUTINE prodc

!=========================================================================!
!                          SUBROUTINE MEANS                               !
!=========================================================================!
! This subroutine computes the mean values mean=(1,...,1)\phi'(1) and     !
! hmean=(1,...,1)\hat\phi'(1) by means of formulae (11) in the paper      !
!=========================================================================!
! Input variables:                                                        !
!   tae  :  pointer associated with the matrix values of \phi_{even}      !
!           at the odd-numbered roots of 1                                !
!   tao  :  pointer associated with the matrix values of \phi_{odd}       !
!           at the odd-numbered roots of 1                                !
!   thae :  pointer associated with the matrix values of                  !
!           \hat\phi_{even} at the roots of 1                             !
!   mean :  pointer associated withthe value \alpha^{(n)} in the          !
!           formula (11)                                                  !
!   hmean: pointer associated with the value \hat\alpha^{(n)} in the      !
!          formula (11)                                                   !
!=========================================================================!
! Output variables                                                        !
!   mean : pointer associated with the value \alpha^{(n+1)} in the        !
!          formula (11)                                                   !
!   hmean: pointer associated with the value \hat\alpha^{(n+1)} in the    !
!          formula (11)                                                   !
!=========================================================================!
! Used interfaces:  schur_interface                                       !
!=========================================================================!
! Used subroutines: solver                                                !
!=========================================================================!

SUBROUTINE means(tae,tao,thae,mean,hmean)
  USE schur_interface, ONLY : solver
  IMPLICIT NONE
  REAL(KIND(0.d0)),POINTER,DIMENSION(:,:,:) :: tae,tao, thae
  REAL(KIND(0.d0)),POINTER,DIMENSION(:)     :: mean,hmean
  REAL(KIND(0.d0)),POINTER,DIMENSION(:,:)   :: w,v
  REAL(KIND(0.d0))                          :: oneh
  INTEGER                                   :: nb, i
  INTRINSIC size
  nb=SIZE(tae,1)
  ALLOCATE(w(nb,nb))
  ALLOCATE(v(nb,1))
  oneh=1.d0/2.d0
  v(:,1)=1.d0-mean
  w(:,:)=-tao(:,:,1)
  loop1 : DO i=1,nb
     w(i,i)=w(i,i)+1.d0
  END DO loop1
  CALL solver(w,v)
  mean=(mean+1.d0-MATMUL(TRANSPOSE(tae(:,:,1)),v(:,1)))*oneh
  hmean=(hmean-MATMUL(TRANSPOSE(thae(:,:,1)),v(:,1)))*oneh
  RETURN
END SUBROUTINE means

!=========================================================================!
!                          SUBROUTINE PMEANS                              !
!=========================================================================!
! This subroutine computes the mean values mean=(1,...,1)\phi'(1) and     !
! hmean=(1,...,1)\hat\phi'(1)                                             !
!=========================================================================!
! Input variables:                                                        !
!   ae :pointer associated with the matrix coefficients of \phi_{even}    !
!   ao :pointer associated with ehe matrix coefficients of \phi_{odd}     !
!   hae:pointer associated with the matrix coefficients of \hat\phi_{even}!
!   hao:pointer associated with the matrix coefficients of \hat\phi_{odd} !
!=========================================================================!
! Output variables                                                        !
!   mean : pointer associated with the value \alpha^{(0)} in the          !
!          formula (11)                                                   !
!   hmean: pointer associated with the value \hat\alpha^{(0)} in the      !
!          formula (11)                                                   !
!=========================================================================!
SUBROUTINE pmeans(ae, ao, hae, hao, mean, hmean)
  IMPLICIT NONE
  REAL(KIND(0.d0)),POINTER,DIMENSION(:,:,:) :: ae,ao, hae, hao
  REAL(KIND(0.d0)),POINTER,DIMENSION(:) :: mean,hmean
  REAL(KIND(0.d0)) :: sp, sd
  INTEGER :: n, nb, j, l, l2
  INTRINSIC size
  nb=SIZE(ae,1)
  n=SIZE(ae,3)
  ALLOCATE(mean(nb))
  ALLOCATE(hmean(nb))
  mean=0.0d0
  hmean=0.0d0
  loop10 : DO j=1,nb
     loop20 : DO l=1,n
        l2=l+l
        sp=SUM(ae(:,j,l))
        sd=SUM(ao(:,j,l))
        mean(j)=mean(j)+sp*(l2-2)+sd*(l2-1)
     END DO loop20
  END DO loop10
  loop30 : DO j=1,nb
     loop40 : DO l=1,n
        l2=l+l
        sp=SUM(hae(:,j,l))
        sd=SUM(hao(:,j,l))
        hmean(j)=hmean(j)+sd*(l2-2)+sp*(l2-1)
     END DO loop40
  END DO loop30
  RETURN
END SUBROUTINE pmeans


!=========================================================================!
!                          SUBROUTINE SOLVER                              !
!=========================================================================!
! This subroutine solves the real linear system x rmat = termr, where     !
! rmat is a square matrix, termr is a matrix                              !
!=========================================================================!
! Input variables:                                                        !
!   rmat  : pointer associated with the matrix rmat                       !
!   termr :  pointer associated with the vector termr                     !
!=========================================================================!
! Output variables:                                                       !
!  termr: the solution x of the system                                    !
!=========================================================================!
! Used interfaces:  lapack90_interfaces                                   !
!=========================================================================!
! Used subroutines: la_gesv                                               !
!=========================================================================!
SUBROUTINE solver(rmat, termr)
  USE lapack90_interfaces
  IMPLICIT NONE
  REAL(KIND(0.d0)), POINTER, DIMENSION(:,:) :: rmat,termr
  INTEGER, POINTER, DIMENSION(:)            :: ipiv
  INTEGER                                   :: info, nb
  INTRINSIC size, transpose
  nb=SIZE(rmat,1)
  ALLOCATE(ipiv(nb))
  rmat=TRANSPOSE(rmat)
  IF(SIZE(termr,2)>1) termr=TRANSPOSE(termr)
  CALL la_gesv(rmat,termr,ipiv,info)
  IF(SIZE(termr,2)>1) termr=TRANSPOSE(termr)
  RETURN
END SUBROUTINE solver



!=========================================================================!
!                          SUBROUTINE SOLVEC                              !
!=========================================================================!
! This subroutine solves the complex linear system x cmat = termc, where  !
! cmat is a square matrix, termc is a matrix                              !
!=========================================================================!
! Input variables:                                                        !
!   cmat  : pointer associated with the matrix cmat                       !
!   termc : pointer associated with the vector termc                      !
!=========================================================================!
! Output variables:                                                       !
!  termc: pointer associated with the solution X of the system            !
!=========================================================================!
! Used interfaces:  lapack90_interfaces                                   !
!=========================================================================!
! Used subroutines: la_gesv                                               !
!=========================================================================!
SUBROUTINE solvec(cmat, termc)    
  USE lapack90_interfaces
  IMPLICIT NONE
  COMPLEX(KIND(0.d0)), POINTER, DIMENSION(:,:) :: cmat,termc
  INTEGER, POINTER, DIMENSION(:)               :: ipiv
  INTEGER                                      :: info, nb
  INTRINSIC size
  nb=SIZE(cmat,1)
  ALLOCATE(ipiv(nb))
  cmat=TRANSPOSE(cmat)
  termc=TRANSPOSE(termc)
  CALL la_gesv(cmat,termc,ipiv,info)
  termc=TRANSPOSE(termc)
  RETURN
END SUBROUTINE solvec




















